/*
 * Combine the geometry from a number of ASCII STL files, and
 * write the combined geometry to standard output in either STL 
 * or VRML 2.0 format.
 *
 * David Moews, 25-VI-2017.
 *
 * This file is in the public domain.
 */

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Get next non-whitespace character. */
int get_nonws(FILE *f)
{
  int c;

  while (((c = getc(f)) != EOF) && isspace(c))
        ;

  return c;
}

/* Require next characters read to be the string s. */
/* Returns 1 on success, 0 on failure. */
int get_string(FILE *f, const char *s)
{
  char c;

  while ((c = *s++) != '\0')
     if (getc(f) != c)
        return 0;

  return 1;
}

/* Read to end of line. */
/* Returns 1 on success, 0 on failure. */
int get_line(FILE *f)
{
  int c;

  while ((c = getc(f)) != EOF)
  {
    if ((c == '\n') || (c == '\r'))
       return 1;
  }

  return 0;
}

struct Triangle
{
  double normal[3];
  double vertices[3][3];
  size_t vertex_num[3];
  double color[3];
};

#define INIT_ALLOC 100

/* Returns 1 on success, 0 on failure. */
int get_ascii_stl(FILE *f, struct Triangle **p_triangles, size_t *p_n_triangles, size_t *p_n_allocated)
{
  assert(*p_n_triangles <= *p_n_allocated);
  if (*p_n_allocated == 0)
     assert(*p_triangles == NULL);
  else
     assert(*p_triangles != NULL);

  if ((get_nonws(f) != 's') || (!get_string(f, "olid")) || (!get_line(f)))
    return 0;

  for (;;)
  {
    int c = get_nonws(f);
    int i;
    struct Triangle t;

    if (c == 'e') 
       break;

    if (c != 'f')
      return 0;

    if ((!get_string(f, "acet")) || (get_nonws(f) != 'n') || 
        (!get_string(f, "ormal")))
      return 0;

    if (fscanf(f, "%lf %lf %lf", &t.normal[0], &t.normal[1], &t.normal[2]) != 3)
      return 0;

    if ((get_nonws(f) != 'o') || (!get_string(f, "uter")) ||
        (get_nonws(f) != 'l') || (!get_string(f, "oop")))
      return 0;

    for (i = 0; i < 3; i++)
    {
      if ((get_nonws(f) != 'v') || (!get_string(f, "ertex")))
        return 0;

      if (fscanf(f, "%lf %lf %lf", &t.vertices[i][0], &t.vertices[i][1], &t.vertices[i][2]) != 3)
        return 0;
    }

    if ((get_nonws(f) != 'e') || (!get_string(f, "ndloop")) ||
        (get_nonws(f) != 'e') || (!get_string(f, "ndfacet")))
      return 0;

    if (*p_triangles == NULL)
    {
      *p_triangles = (struct Triangle *)malloc(INIT_ALLOC * sizeof(struct Triangle));
      if (*p_triangles == NULL)
      {
        fprintf(stderr, "Error: Out of memory.\n");
        exit(1);
      }
      *p_n_allocated = INIT_ALLOC;
    }

    if (*p_n_triangles == *p_n_allocated)
    {
      *p_n_allocated *= 2;
      *p_triangles = 
        (struct Triangle *)realloc(*p_triangles, *p_n_allocated * sizeof(struct Triangle));
      if (*p_triangles == NULL)
      {
        fprintf(stderr, "Error: Out of memory.\n");
        exit(1);
      }
    }

    assert(*p_n_triangles < *p_n_allocated);

    (*p_triangles)[(*p_n_triangles)++] = t;
  }

  if (!get_string(f, "ndsolid"))
    return 0;

  return 1;
}

#undef INIT_ALLOC

struct Color_record
{
  const char *name;
  const double val[3];
};

static const struct Color_record color_rec[] =  
{ 
  {"red", {1, 0, 0}},
  {"green", {0, 1, 0}},
  {"blue", {0, 0, 1}},
  {"yellow", {1, 1, 0}},
  {"gray", {0.5, 0.5, 0.5}},
  {"white", {1, 1, 1}},
  {"black", {0, 0, 0}}
};
     
int get_color(const char *s, double color[3])
{
  unsigned i, j;

  for (i = 0; i < sizeof(color_rec) / sizeof(struct Color_record); i++)
      if (strcmp(s, color_rec[i].name) == 0)
      {
        for (j = 0; j < 3; j++)
            color[j] = color_rec[i].val[j];

        return 1;
      }

  return 0;
}

void usage(const char *nm)
{
  int i = 0;

  fprintf(stderr, "\nUsage: %s [input-stl-file | option] ...\n\
\n\
Options:\n\
\n\
  --stl            Output combined geometry to stdout in STL format.\n\
  --vrml2.0        Output combined geometry to stdout in VRML 2.0 format.\n\
  --color <name>   Succeeding files should be output in the given color.\n\
                   Possible color names: ", nm);
  for (;;) 
  {
    fprintf(stderr, "'%s'", color_rec[i].name);
    i++;
    if (i == sizeof(color_rec) / sizeof(struct Color_record))
        break;
    fputs(", ", stderr);
  }
  fputs(".\n", stderr);
}

struct Vertex_record
{
  double coords[3];
  size_t triangle_num;
  int vertex_num; 
};

int compare_vertex_records(const void *a, const void *b)
{
  const struct Vertex_record *ap = a;
  const struct Vertex_record *bp = b;
  int i;

  for (i = 0; i < 3; i++)
  {
    if (ap->coords[i] < bp->coords[i])
       return -1;

    if (ap->coords[i] > bp->coords[i])
       return 1;
  }

  if (ap->triangle_num < bp->triangle_num)
     return -1;

  if (ap->triangle_num > bp->triangle_num)
     return 1;

  if (ap->vertex_num < bp->vertex_num)
     return -1;

  if (ap->vertex_num > bp->vertex_num)
     return 1;

  return 0;
}

int main(int argc, char **argv)
{
  double current_color[3] = {1, 1, 1};
  double last_vertex_coords[3];
  size_t last_vertex_idx;
  struct Triangle *triangles = NULL;
  size_t n_triangles = 0;
  size_t n_allocated = 0;
  size_t it;
  size_t it_out;
  int i;
  int j;
  int output_format = -1;

  for (i = 1; i < argc; i++)
      if (argv[i][0] == '-')
      {
        if (strcmp(argv[i], "--stl") == 0)
        {
          if ((output_format >= 0) && (output_format != 0))
          {
            fprintf(stderr, "Error: Conflicting output formats.\n");
            usage(argv[0]);
            exit(6);
          }

          output_format = 0;
          continue;
        }

        if (strcmp(argv[i], "--vrml2.0") == 0)
        {
          if ((output_format >= 0) && (output_format != 1))
          {
            fprintf(stderr, "Error: Conflicting output formats.\n");
            usage(argv[0]);
            exit(6);
          }

          output_format = 1;
          continue;
        }

        if (strcmp(argv[i], "--color") != 0)
        {
          fprintf(stderr, "Error: Invalid option '%s'.\n", argv[i]);
          usage(argv[0]);
          exit(2);
        }

        if (++i == argc)
        {
          fprintf(stderr, "Error: No color after '--color'.\n");
          usage(argv[0]);
          exit(3);
        }

        if (!get_color(argv[i], current_color))
        {
          fprintf(stderr, "Error: Bad color '%s' after '--color'.\n", argv[i]);
          usage(argv[0]);
          exit(4);
        }
      }
      else
      {
        size_t n_triangles_initial = n_triangles;
        FILE *f = fopen(argv[i], "r");

        if (f == NULL)
        {
          fprintf(stderr, "Error: Can't open file '%s'.\n", argv[i]);
          exit(7);
        }

        if (!get_ascii_stl(f, &triangles, &n_triangles, &n_allocated))
        {
          fprintf(stderr, "Error: Bad STL file '%s'.\n", argv[i]);
          exit(5);
        }

        fclose(f);

        for (it = n_triangles_initial; it < n_triangles; it++)
            for (j = 0; j < 3; j++)
                triangles[it].color[j] = current_color[j];
      }

  if (output_format < 0)
     output_format = 0;

  if (output_format == 0)
  {
    fputs("solid default\n", stdout);

    for (it = 0; it < n_triangles; it++)
    {
      printf("  facet normal %.10e %.10e %.10e\n", 
             triangles[it].normal[0], triangles[it].normal[1], triangles[it].normal[2]);
      printf("    outer loop\n");
      for (i = 0; i < 3; i++)
          printf("      vertex %.10e %.10e %.10e\n",
                 triangles[it].vertices[i][0],
                 triangles[it].vertices[i][1],
                 triangles[it].vertices[i][2]);
      printf("    endloop\n");
      printf("  endfacet\n");
    }
    fputs("endsolid default\n", stdout);

    return 0;
  }

  size_t n_vertices = n_triangles * 3;

  struct Vertex_record *vertex_records = 
          (struct Vertex_record *)malloc(n_vertices * sizeof(struct Vertex_record));
  if (vertex_records == NULL)
  {
    fprintf(stderr, "Error: Out of memory.\n");
    exit(1);
  }

  for (it = 0; it < n_triangles; it++)
      for (i = 0; i < 3; i++)
      {
        for (j = 0; j < 3; j++)
            vertex_records[it * 3 + i].coords[j] = triangles[it].vertices[i][j];
        vertex_records[it * 3 + i].triangle_num = it;
        vertex_records[it * 3 + i].vertex_num = i;
      }

  qsort(vertex_records, n_vertices, sizeof(struct Vertex_record), compare_vertex_records);

  printf("#VRML V2.0 utf8\nShape {\ngeometry IndexedFaceSet {\ncoord Coordinate {\npoint [\n");

  it_out = 0;

  for (it = 0; it < n_vertices; it++)
  {
    if ((it == 0) || 
        ((vertex_records[it].coords[0] != last_vertex_coords[0]) ||
         (vertex_records[it].coords[1] != last_vertex_coords[1]) ||
         (vertex_records[it].coords[2] != last_vertex_coords[2])))
    {
      for (i = 0; i < 3; i++)
          last_vertex_coords[i] = vertex_records[it].coords[i];
      printf("%.10e %.10e %.10e\n", 
             last_vertex_coords[0], last_vertex_coords[1], last_vertex_coords[2]);
      last_vertex_idx = it_out++;
    }
    triangles[vertex_records[it].triangle_num].
             vertex_num[vertex_records[it].vertex_num] = last_vertex_idx;
  }

  printf("] }\ncolorPerVertex FALSE\ncolor Color {\ncolor [\n");

  for (it = 0; it < n_triangles; it++)
      printf("%.10e %.10e %.10e\n", 
             triangles[it].color[0], triangles[it].color[1], triangles[it].color[2]);

  printf("] }\ncoordIndex [\n");

  for (it = 0; it < n_triangles; it++)
      printf("%zu %zu %zu -1\n", 
             triangles[it].vertex_num[0], 
             triangles[it].vertex_num[1], 
             triangles[it].vertex_num[2]);       

  printf("]\n}\n}\n");

  return 0;
}
