//  OpenSCAD file with basic functions for constructing slide adder.
//
//  David Moews and Paul C. Moews, 18-VI-2017 to 2-VII-2017.
//  
//  This file is in the public domain.


//
// Basic definitions.
//

digit_string = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

base = 10;

num_places = 5;

initial_entry = [10, 5, 0, 9, 7];

//
// Font to use for text.
//

textfont = "Arial Black";

//
// Clearance between sliding parts.
//

clearance = 0.425;

//
// Thickness of embossed characters.
//

embossed_char_depth = 1;

//
// Strip dimensions.
//

depth = 2.5;

tooth_height = 3.4;

gap_height = 3.4;

strip_width = 5.25;

tooth_width = 2.7;

// Scale of embossed character height relative to baseline spacing.

strip_character_scale = 0.9;

//
// Tray dimensions.
//

// Width of separator between strips 
//    (also used for border of tray and tabs, and lid overhang.)

separator_width = 2;

// Thickness of bottom of tray.

tray_depth = 3;

// Thickness of separators in tray between strips, and rim of tray.

separator_depth = 3.1;

//
// Lid dimensions.
//

lid_depth = 2;

// Scale of embossed character height relative to baseline spacing.

lid_character_scale = 0.9;

// Scale of digit-displaying hole relative to baseline spacing.

digit_hole_scale = 1.0;

//
// Further dimensions computed from the above.
//

corner_radius = min(gap_height, tooth_width);

column_delta = separator_width + strip_width + 2 * tooth_width + 2 * clearance;
 
step_height = tooth_height + gap_height;
 
// Size and origin of working area.

x_origin = - (strip_width / 2 + tooth_width + clearance);
y_origin = - (base * step_height + clearance);

x_extent = (num_places - 1) * column_delta + strip_width + 2 * tooth_width + 2 * clearance;
y_extent = step_height * (3 * base + 2) + 2 * clearance;

// Tabs.

tab_positions = [[x_origin - separator_width, -step_height, 0],
                 [x_origin + x_extent, (base + 2.5) * step_height, 0],
                 [column_delta, y_origin + y_extent, 0],
                 [(num_places - 2) * column_delta, y_origin - separator_width, 0]];

tab_orientations = [0, 0, 1, 1];

module strip_digits()
{
    for (i = [0 : base - 1])
        translate(v = [tooth_width / 2, 
                       i * step_height + (1 - strip_character_scale) * step_height / 2, 
                       depth])
            linear_extrude(embossed_char_depth)
                text(text = digit_string[i], size = strip_character_scale * step_height, 
                     halign = "center", font = textfont);
    
    translate(v = [tooth_width / 2, 
                   base * step_height + (1 - strip_character_scale) * step_height / 2, 
                   depth])
        linear_extrude(embossed_char_depth)
            text(text = "+", size = strip_character_scale * step_height, halign = "center",
                 font = textfont);
}

module strip()
{
    difference()
    { 
        union()
        {
            translate(v = [-strip_width / 2 - tooth_width, 0, 0])
                cube([2 * tooth_width + strip_width, (base + 2) * step_height, depth]);

            translate(v = [-strip_width / 2, 0, 0])
                cube([strip_width, (2 * base + 2) * step_height, depth]);
        }

        for (i = [2 : base + 1])
        {
            translate(v = [-strip_width / 2 - tooth_width / 2 + clearance, 
                           gap_height / 2 + i * step_height, -clearance])
                    linear_extrude(depth + 2 * clearance)
                        polygon([[-tooth_width / 2, -gap_height / 2],
                                 [ tooth_width / 2, -gap_height / 2],
                                 [ 0, gap_height / 2]]);
        }
    }

    for (i = [base + 2 : 2 * base + 1])
    {
        translate(v = [0, gap_height + i * step_height, 0])
        {
            if (i == 2 * base + 1)
            {
                translate(v = [-(tooth_width + strip_width / 2), 0, 0])
                    cube([strip_width + tooth_width, tooth_height, depth]);
            } else
            {
                translate(v = [-(tooth_width + strip_width / 2), 0, 0])
                    cube([strip_width + 2 * tooth_width, tooth_height, depth]);
            }
        }
    }

    strip_digits();
}

module strip_set()
{
    for (j = [0 : num_places - 1])
        translate(v = [j * column_delta, -step_height * initial_entry[j], 0])
            strip();
}

module separator_set()
{
    for (j = [0 : num_places - 2])
        translate(v = [j * column_delta + strip_width / 2 + tooth_width + clearance, y_origin, 0])
        {
            cube([separator_width, (2 * base + 2) * step_height + clearance, 
                  separator_depth + clearance]);

            translate(v = [0, (2 * base + 3) * step_height + clearance, 0])
                cube([separator_width, (base - 1) * step_height + clearance, 
                      separator_depth + clearance]);
        }
}

module tray()
{
    translate(v = [x_origin, y_origin, 0])
    {
        difference()
        {
            translate(v = [-separator_width, -separator_width, 0])
                cube([2 * separator_width + x_extent, 2 * separator_width + y_extent, 
                      separator_depth + clearance]);

            cube([x_extent, y_extent, separator_depth + 2 * clearance]);
        }
        translate(v = [-separator_width, -separator_width, -tray_depth])
            cube([2 * separator_width + x_extent, 2 * separator_width + y_extent, tray_depth]);
    }

    for (j = [0 : len(tab_positions) - 1])
    {
        translate(v = tab_positions[j])
        {
            if (tab_orientations[j] == 0)
                cube([separator_width, step_height, 
                      2 * clearance + separator_depth + lid_depth * 2]);
            else
                cube([step_height, separator_width, 
                      2 * clearance + separator_depth + lid_depth * 2]);
        }
    }
    
    separator_set();
}

module lid_digits()
{
    for (i = [0 : base - 1])
        translate(v = [0, step_height * (i + 2) + (1 - lid_character_scale) * step_height / 2,
                       lid_depth])
            linear_extrude(embossed_char_depth)
                text(text = digit_string[i], size = lid_character_scale * step_height, 
                     halign = "center", font = textfont);
}

module lid_hole(carry_hole)
{
    translate(v = [tooth_width / 2, step_height / 2, lid_depth / 2])
        cube([digit_hole_scale * (strip_width + tooth_width), 
              digit_hole_scale * step_height, lid_depth + 2 * clearance], 
             center = true);

    translate(v = [-(strip_width / 2 + tooth_width + clearance + separator_width / 2),
                   2 * step_height, -clearance])
    {
        difference()
        {
            union()
            {
                cube([separator_width / 2 + clearance + tooth_width, 
                      step_height * base + gap_height, lid_depth + 2 * clearance]);

                if (carry_hole)
                   translate(v = [-(tooth_width + clearance + separator_width / 2),
                                  step_height * (base - 1), 0])
                       cube([separator_width + clearance * 2 + tooth_width * 2,
                             step_height + gap_height, lid_depth + 2 * clearance]);
            }

            translate(v = [separator_width / 2 + clearance + tooth_width - corner_radius,
                           step_height * base + gap_height - corner_radius, -clearance])
                difference()
                {
                    cube([corner_radius + clearance, corner_radius + clearance, 
                          lid_depth + 4 * clearance]);

                    translate(v = [0, 0, -clearance])
                        cylinder(r = corner_radius, h = lid_depth + 6 * clearance);
                }

            if (carry_hole)
                translate(v = [-(tooth_width + 2 * clearance + separator_width / 2),
                       step_height * base + gap_height - corner_radius, -clearance])
                    difference()
                    {
                        cube([corner_radius + clearance, corner_radius + clearance, 
                              lid_depth + 4 * clearance]);

                        translate(v = [corner_radius + clearance, 0, -clearance])
                             cylinder(r = corner_radius, h = lid_depth + 6 * clearance);
                    }
        }
    }
}

module lid()
{
 
    difference()
    {  
        translate(v = [x_origin - 2 * separator_width, y_origin - 2 * separator_width, 0])
            cube([4 * separator_width + x_extent, 4 * separator_width + y_extent, lid_depth]);  

        lid_hole(false);

        for (j = [1 : num_places - 1])
            translate(v = [j * column_delta, 0, 0])
                 lid_hole(true);

        for (j = [0 : len(tab_positions) - 1])
        {
            translate(v = tab_positions[j])
                translate(v = [-clearance, -clearance, -clearance])
                {
                    if (tab_orientations[j] == 0)
                        cube([separator_width + 2 * clearance, step_height + 2 * clearance, 
                              lid_depth + 2 * clearance]);
                    else
                        cube([step_height + 2 * clearance, separator_width + 2 * clearance, 
                              lid_depth + 2 * clearance]);
                }
        }
    }

    for (j = [0 : num_places - 1])
        translate(v = [j * column_delta, 0, 0])
           lid_digits();        
}
